/*
 * include/nfshell.h
 *
 * Common definitions for nfshell
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#ifndef NFSBENCH_H
#define NFSBENCH_H

#include <sys/types.h>
#include <netinet/in.h>
#include <system.h>

#if SIZEOF_UNSIGNED_LONG == 4
typedef unsigned long	u32;
#elif SIZEOF_UNSIGNED_INT == 4
typedef unsigned int	u32;
#else
# error "Don't know how to define unsigned 32-bit type"
#endif

typedef struct sockaddr_in	sockaddr_in;

#define MAXRWSIZE	(16 * 1024)
#define MAXPKTSIZE	(MAXRWSIZE + 1024)
#define MAXBIODS	32
#define MAXSERVS	32
#define MAXREQUESTS	(2 * MAXBIODS)
#define MAXREXMITS	16
#define INITIMEO	(6 * HZ / 10)		/* 0.6 seconds */
#define DEFAULT_UID	5555
#define DEFAULT_GID	5555

struct rpcreq;
struct svcreq;
typedef void		(*rqcallback_t)(struct rpcreq *req);
typedef u32 *		(*rqxdrproc_t)(struct rpcreq *, u32 *, void *);
typedef u32 *		(*svxdrproc_t)(struct svcreq *, u32 *, void *);
typedef u32 *		(*svcdspproc_t)(struct svcreq *, u32 *, void *);

struct rpcprog {
	sockaddr_in		pg_server;
	u32			pg_prog;
	u32			pg_vers;
	rqxdrproc_t *		pg_encode;
	rqxdrproc_t *		pg_decode;
};

struct rpcserv {
	int			sv_sock;
	char *			sv_name;
	u32			sv_prog;
	u32			sv_vers;
	u32			sv_port;
	struct rpcserv *	sv_next;
	svcdspproc_t *		sv_func;
	int			sv_nrprocs;
	rqxdrproc_t *		sv_decode;
	rqxdrproc_t *		sv_encode;
	void *			sv_data;
};

struct rpcreq {
	struct rpcreq *		rq_next;
	sockaddr_in *		rq_addr;
	int			rq_sock;
	u32			rq_xid;
	u32 *			rq_msg;
	unsigned int		rq_msglen;
	u32 *			rq_rply;
	unsigned int		rq_rplylen;
	unsigned long		rq_qtime;
	unsigned long		rq_rtt;
	unsigned long		rq_expires;
	unsigned long		rq_timeout;
	unsigned short		rq_rexmits;
	rqxdrproc_t		rq_encode;
	rqxdrproc_t		rq_decode;
	rqcallback_t		rq_callback;
	void *			rq_clntdata;
	void *			rq_resp;
	u32			rq_rpcstat;
};

struct svcreq {
	sockaddr_in *		rq_addr;
	u32			rq_xid;
	u32			rq_prog;
	u32			rq_vers;
	u32			rq_proc;
	u32 *			rq_msg;
	unsigned int		rq_msglen;
	u32 *			rq_rply;
	unsigned int		rq_rplylen;
	u32			rq_rpcstat;
};

struct rpcstat {
	int			pkt_sent;
	int			pkt_recv;
	int			pkt_lost;
	int			pkt_rxmt;
	int			byte_sent;
	int			byte_recv;

	unsigned long		start_time;
	unsigned long		total_time;
	unsigned long		rtt;
	unsigned long		min_rtt;
	unsigned long		max_rtt;
};

extern struct rpcreq *		req_pending;
extern struct rpcreq *		req_pending_tail;
extern struct rpcreq *		req_queued;
extern struct rpcreq *		req_queued_tail;
extern struct rpcreq *		req_free;
extern struct rpcstat		stats;

/*
 * Configuration variables.
 */
extern int			nrbiods;
extern int			nrsocks;
/*
extern int			nfs_rsize;
extern int			nfs_wsize;
extern char *			nfs_root;
extern sockaddr_in		nfs_server;
extern struct in_addr		nfs_addr;
extern unsigned short		nfs_port;
 */

#include "xdr/rpc.h"
#include "xdr/nfs.h"

/* Approximation of current time in jiffies */
extern unsigned long		xtime;

/*
 * Functions
 */
struct rpcprog *rpc_clntcreate(sockaddr_in *, u32, u32, u32);
void		rpc_clntdestroy(struct rpcprog *);
struct rpcserv *rpc_svccreate(u32, u32);
void		rpc_svcdestroy(struct rpcserv *);
void		rpc_svcdispatch(int, u32 *, int, sockaddr_in *);
int		rpc_call(struct rpcprog *, int, void *, void *);
int		rpc_setuid(int);
int		rpc_setgid(int);

void		biod(int *);
void		biod_drain(void);
void		biod_drain_on(void);
void		biod_flush(void);
void		biod_svc_check(void);
int		biod_svc_sock(u32 *portp);

struct rpcreq *	generate_request(sockaddr_in *, rqcallback_t, void *);
void		exec_request(struct rpcreq *);
void		enqueue_request(struct rpcreq *);
void		insert_request(struct rpcreq **, struct rpcreq *);
void		finish_request(struct rpcreq *, u32 *, unsigned int);

unsigned long	jiffies(void);
char *		jiffies_print(unsigned long time);
char *		jiffies_print_100(unsigned long time);
int		addr_parse(char *hname, struct in_addr *ap);
char *		addr_print(struct in_addr *ap);
int		fh_parse(char *, struct nfs_fh *);
char *		fh_print(struct nfs_fh *);
u32		rpcsvc_parse(char *name);
void *		xmalloc_fl(unsigned long, char *, int);
char *		xstrdup(char *sp);
void		xfree(void *);
void		clear_stats(void);
void		print_stats(void);

#define append_request(head, tail, req)	\
	req->rq_next = NULL;					\
	if (!tail) {						\
		head = tail = req;				\
	} else if (tail->rq_expires < (req)->rq_expires) {	\
		tail->rq_next = req;				\
		tail = req;					\
	} else {						\
		insert_request(&head, req);			\
		if (!(req)->rq_next)				\
			tail = (req);				\
	}
#define wait_request(req)		\
	append_request(req_pending, req_pending_tail, req)
#define xmalloc(size)			xmalloc_fl(size, __FILE__, __LINE__)
#define biod_sync(statement)		\
	{							\
		biod_flush();					\
		statement;					\
		biod_flush();					\
	}
#define exec_request(req)		biod_sync(enqueue_request(req))

#define script_nextaction()	(0)

#endif /* NFSBENCH_H */
