/*
 * shell/args.h
 *
 * Generic TCL argument parser for NFS TCL binding
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#ifndef SHELL_ARGS_H
#define SHELL_ARGS_H

#include <tcl.h>
#include <nfshell.h>
#include <macros.h>

/*
 * Type definition
 */
struct tcl_type_info {
	char *			name;
	void			(*init)(void *, caddr_t);
	int			(*print)(Tcl_Interp *, void *, caddr_t);
	int			(*parse)(Tcl_Interp *, void *, caddr_t, char *);
	void *			data;
};

/*
 * This struct decribes a single argument
 */
struct tcl_arg_info {
	char *			member;
	unsigned int		offset;
	struct tcl_type_info *	type;
};

/*
 * This is a set of arguments grouped together into a struct
 */
struct tcl_parm_info {
	char *			name;
	struct tcl_arg_info *	args;
	unsigned int		nrargs;
};

#define tcl_init_type(type, ptr)					\
	do {								\
		if ((type)->init)					\
			(type)->init((type)->data, ptr);		\
	} while (0)

#define tcl_print_type(interp, type, ptr)				\
	(type)->print(interp, (type)->data, ptr)

extern int  tcl_parse_args(Tcl_Interp *, struct tcl_type_info *,
				caddr_t, int, char **);
extern int  tcl_parse_type(Tcl_Interp *, struct tcl_type_info *,
				caddr_t, char *);
extern int  tcl_set_result(Tcl_Interp *, struct tcl_type_info *, caddr_t);
extern void tcl_struct_init(struct tcl_parm_info *, caddr_t);
extern int  tcl_struct_parse(Tcl_Interp *, struct tcl_parm_info *,
				caddr_t, char *);
extern int  tcl_struct_print(Tcl_Interp *, struct tcl_parm_info *, caddr_t);
extern int  tcl_arginfo(Tcl_Interp *, char *, struct tcl_type_info *);
extern int  tcl_return(Tcl_Interp *, char *, ...);
extern int  tcl_error(Tcl_Interp *, char *, ...);

/*
 * The following looks extremely ugly, but is necessary to bludgeon cpp
 * into doing what I want:
 */
#define TCL_TYPENAME(tname)	CONCAT3(tcl_, tname, _type)
#define TCL_PARMNAME(sname)	STRING(sname)
#define TCL_MEMBNAME(sname)	CONCAT3(tcl_, sname, _members)
#define TCL_STRUCTNAME(sname)	CONCAT2(tcl_, sname)
#define TCL_INIT_FUNC(sname)	CONCAT3(tcl_, sname, _init)
#define TCL_PRINT_FUNC(sname)	CONCAT3(tcl_, sname, _print)
#define TCL_PARSE_FUNC(sname)	CONCAT3(tcl_, sname, _parse)
#define TCL_TYPE_static		static
#define TCL_TYPE_extern

#define __TCL_TYPE_STRUCT(qual, tname, init, print, parse, data)	\
TCL_TYPE_##qual struct tcl_type_info	TCL_TYPENAME(tname) = {		\
	STRING(tname),							\
	(void (*)(void *, caddr_t)) init,				\
	(int (*)(Tcl_Interp *, void *, caddr_t)) print,			\
	(int (*)(Tcl_Interp *, void *, caddr_t, char *)) parse,		\
	data								\
}

#define TCL_DECLARE_TYPE(qual, tname)					\
qual struct tcl_type_info	TCL_TYPENAME(tname)

#define TCL_DEFINE_TYPE(qual, tname, data)				\
__TCL_TYPE_STRUCT(qual, tname, TCL_INIT_FUNC(tname),			\
		TCL_PRINT_FUNC(tname), TCL_PARSE_FUNC(tname), data)

#define TCL_DEFINE_STRUCT(qual, sname)					\
__TCL_TYPE_STRUCT(qual, sname, TCL_INIT_FUNC(sname), tcl_struct_print, 	\
		tcl_struct_parse, &TCL_STRUCTNAME(sname))

#define TCL_BEGIN_STRUCT(qual, sname)					\
static struct tcl_arg_info	TCL_MEMBNAME(sname)[] = {

#define TCL_MEMBER(sname, type, memb)					\
   { STRING(memb), memoffset(sname, memb), &TCL_TYPENAME(type) }

#define TCL_END_STRUCT(qual, sname)					\
};									\
									\
static struct tcl_parm_info	TCL_STRUCTNAME(sname) = {		\
     TCL_PARMNAME(sname), TCL_MEMBNAME(sname),				\
     numberof(TCL_MEMBNAME(sname))					\
};

/*
#define TCL_PARSE(sname) \
   tcl_parse_args(interp, &TCL_STRUCTNAME(sname), (caddr_t) &args, argc, argv)

#define TCL_PRINT(sname) \
   tcl_print_result(interp, &TCL_STRUCTNAME(sname), (caddr_t) &res)
 */

/*
 * Specific low-level types
 */
typedef char *		string;

TCL_DECLARE_TYPE(extern, void);
TCL_DECLARE_TYPE(extern, int);
TCL_DECLARE_TYPE(extern, u32);
TCL_DECLARE_TYPE(extern, timeval);
TCL_DECLARE_TYPE(extern, string);

/*
 * TCL procedure registration
 */
#define TCL_PROT_REGISTER(prot)		CONCAT3(tcl_, prot, _register)

#endif /* SHELL_ARGS_H */
