/*
 * rpc/client.c
 *
 * Handling of RPC clients and requests
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <nfshell.h>
#include <xdr/rpc.h>
#include <xdr/portmap.h>
#include <xdr/mount.h>
#include <xdr/nfs.h>
#include <xdr/nlm.h>

/*
 * Create an RPC program handle
 */
struct rpcprog *
rpc_clntcreate(sockaddr_in *sin, u32 prog, u32 vers, u32 port)
{
	struct rpcprog	*pg;
	rqxdrproc_t	*enctab, *dectab;

	switch (prog) {
	case PMAP_PROGRAM:
		enctab = pmap_encode;
		dectab = pmap_decode;
		break;
	case MOUNT_PROGRAM:
		enctab = mnt_encode;
		dectab = mnt_decode;
		break;
	case NFS_PROGRAM:
		enctab = nfs_encode;
		dectab = nfs_decode;
		break;
	case NLM_PROGRAM:
		enctab = nlm_encode;
		dectab = nlm_decode;
		break;
	default:
		return NULL;
	}

	if (!(pg = (struct rpcprog *) xmalloc(sizeof(*pg))))
		return NULL;

	pg->pg_server = *sin;
	pg->pg_prog   = prog;
	pg->pg_vers   = vers;
	pg->pg_encode = enctab;
	pg->pg_decode = dectab;

	if (sin->sin_port == 0)
		pg->pg_server.sin_port = htons((unsigned short) port);

	return pg;
}

/*
 * Destroy an RPC program handle
 */
void
rpc_clntdestroy(struct rpcprog *pg)
{
	xfree(pg);
}

static void
rpc_callback(struct rpcreq *req)
{
	rqxdrproc_t	decode = req->rq_decode;
	u32		*msg = req->rq_rply,
			*statp = (u32 *) req->rq_clntdata;

	if (rpc_decode_reply(req, &msg) != 0) {
		msg = decode(req, msg, req->rq_resp);
		if (!xdr_check_rplylen(req, msg))
			req->rq_rpcstat = RPC_CANTDECODERES;
	}
	*statp = req->rq_rpcstat;
}

int
rpc_call(struct rpcprog *prog, int proc, void *argp, void *resp)
{
	struct rpcreq	*req;
	u32		*msg, rpcstat;

	req = generate_request(&prog->pg_server, rpc_callback, &rpcstat);
	req->rq_decode = prog->pg_decode[proc];
	req->rq_resp = resp;

	msg = rpc_encode_header(req->rq_msg, prog->pg_prog,
						prog->pg_vers, proc);

	if ((msg = (prog->pg_encode[proc])(req, msg, argp)) == NULL)
		return RPC_CANTENCODEARGS;
	xdr_adjust_msglen(req, msg);

	exec_request(req);
	return (int) rpcstat;
}

int
rpc_setuid(int uid)
{
	int	olduid = auth_uid;

	auth_uid = uid;
	return olduid;
}

int
rpc_setgid(int gid)
{
	int	oldgid = auth_gid;

	auth_gid = gid;
	return oldgid;
}
