/*
 * rpc/server.c
 *
 * Handling of RPC sevices
 *
 * Copyright (C) 1996, 1997 Olaf Kirch <okir@monad.swb.de>
 */

#include <memory.h>
#include <nfshell.h>
#include <macros.h>
#include <xdr/rpc.h>
#include <xdr/portmap.h>
#include <xdr/mount.h>
#include <xdr/nfs.h>
#include <xdr/null.h>

static struct rpcserv *	services = NULL;

/*
 * Table of supported services
 */
#define RPC_SERVICE(NAME, name)	\
  {	STRING(name),							\
	CONCAT2(NAME, _PROGRAM),	CONCAT2(NAME, _VERSION),	\
	CONCAT3(svc_, name, _dispatch),	CONCAT2(NAME, _MAXPROC) + 1	\
  }
static struct {
	char *		name;
	u32		prog;
	u32		vers;
	svcdspproc_t *	func;
	unsigned int	nprocs;
}			svc_services[] = {
	RPC_SERVICE(NUL, nul),
  	{ NULL, }
};

/*
 * Create an RPC program handle
 */
struct rpcserv *
rpc_svccreate(u32 prog, u32 vers)
{
	struct rpcserv	*serv;
	svcdspproc_t	*svcfunc = NULL;
	char		*svcname = NULL;
	int		svcprocs = 0, i;

	for (serv = services; serv; serv = serv->sv_next) {
		if (serv->sv_prog == prog && serv->sv_vers == vers)
			return NULL;
	}

	for (i = 0; svc_services[i].name; i++) {
		if (svc_services[i].prog == prog
		 && svc_services[i].vers == vers) {
			svcfunc  = svc_services[i].func;
			svcname  = svc_services[i].name;
			svcprocs = svc_services[i].nprocs;
			break;
		}
	}
	if (!svcfunc)
		return NULL;

	if (!(serv = (struct rpcserv *) xmalloc(sizeof(*serv))))
		return NULL;

	memset(serv, 0, sizeof(*serv));
	serv->sv_sock    = biod_svc_sock(&serv->sv_port);
	serv->sv_name    = svcname;
	serv->sv_prog    = prog;
	serv->sv_vers    = vers;
	serv->sv_nrprocs = svcprocs;
	serv->sv_func    = svcfunc;
	serv->sv_data    = NULL;
	serv->sv_next    = services;
	services = serv;

	return serv;
}

/*
 * Destroy an RPC program handle
 */
void
rpc_svcdestroy(struct rpcserv *serv)
{
	struct rpcserv	**spp;

	for (spp = &services; *spp; spp = &(*spp)->sv_next) {
		if (*spp == serv) {
			*spp = serv->sv_next;
			xfree(serv);
			return;
		}
	}
}

void
rpc_svcdispatch(int fd, u32 *buffer, int len, sockaddr_in *sin)
{
	struct svcreq	req;
	struct rpcserv	*serv;
	u32		rbuffer[MAXPKTSIZE/4], *msg, *statp;
	u32		low = ~0, high = 0;

	/* Datagram must at least have direction == CALL */
	if (len < 40 || buffer[1] != 0)
		return;

	memset(&req, 0, sizeof(req));
	req.rq_addr	= sin;
	req.rq_msg	= buffer;
	req.rq_msglen	= (unsigned int) len;
	req.rq_rply	= rbuffer;
	req.rq_rplylen	= MAXPKTSIZE;
	req.rq_xid	= buffer[0];

	if (!rpc_decode_header(&req, &msg)) {
		msg = rpc_encode_reply(&req, req.rq_rply, 0);
		statp = msg - 1;

		req.rq_rpcstat = 1;		/* PROG_UNAVAIL */
		for (serv = services; serv; serv = serv->sv_next) {
			if (serv->sv_prog != req.rq_prog)
				continue;
			req.rq_rpcstat = 2;	/* PROG_MISMATCH */
			if (serv->sv_vers == req.rq_vers) {
				if (req.rq_proc < serv->sv_nrprocs) {
					req.rq_rpcstat = 0; /* Okay for now */
				} else {
					req.rq_rpcstat = 3; /* PROC_UNAVAIL */
				}
				break;
			}
			if (serv->sv_vers < low)
				low = serv->sv_vers;
			if (serv->sv_vers > high)
				high = serv->sv_vers;
		}

		if (req.rq_rpcstat == 0) {
			(serv->sv_func[req.rq_proc])(&req, msg, serv->sv_data);
		}
		if (req.rq_rpcstat != 0) {
			*statp = htonl(req.rq_rpcstat);
			msg = statp + 1;
			if (req.rq_rpcstat == 2 && req.rq_vers == 0) {
				/* Client query for available versions */
				*msg++ = htonl(low);
				*msg++ = htonl(high);
			}
		}
	}
	xdr_adjust_rplylen(&req, msg);

	sendto(fd, rbuffer, req.rq_rplylen, 0,
			(struct sockaddr *) sin, sizeof(*sin));
	return;
}
