/*
 * shell/args.c
 *
 * Generic argument parser for NFS TCL binding
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <malloc.h>
#include <shell/args.h>

static int	common_struct_parse(Tcl_Interp *, struct tcl_parm_info *,
					caddr_t, int, char **, int);
static int	tcl_struct_help(Tcl_Interp *, char *, struct tcl_parm_info *);
static int	tcl_type_help(Tcl_Interp *, char *, struct tcl_type_info *);

int
tcl_parse_args(Tcl_Interp *interp, struct tcl_type_info *type,
					caddr_t ptr, int argc, char **argv)
{
	char	*proc = argv[0];
	int	error;

	argc -= 2;
	argv += 2;

	if ((void *) type->parse == (void *) tcl_struct_parse) {
		struct tcl_parm_info	*info;

		info = (struct tcl_parm_info *) type->data;

		if (argc == 1 && !strcmp(argv[0], "-help"))
			return tcl_struct_help(interp, proc, info);

		if (info->nrargs != argc)
			return tcl_error(interp,
				"%s: bad number of arguments\n", proc);

		error = common_struct_parse(interp, info, ptr,
							argc, argv, 0);
	} else if (argc == 1) {
		if (!strcmp(argv[0], "-help"))
			return tcl_type_help(interp, proc, type);

		error = type->parse(interp, type->data, ptr, argv[0]);
	} else {
		return tcl_error(interp, "%s: too many arguments", proc);
	}

	return error;
}

int
tcl_struct_parse(Tcl_Interp *interp, struct tcl_parm_info *info,
				caddr_t ptr, char *list)
{
	int	argc, error;
	char	**argv;

	if ((error = Tcl_SplitList(interp, list, &argc, &argv)) != TCL_OK)
		return tcl_error(interp,
				"failed to split %s args list", info->name);

	error = common_struct_parse(interp, info, ptr, argc, argv, 1);

	free(argv);
	return error;
}

static int
common_struct_parse(Tcl_Interp *interp, struct tcl_parm_info *info,
			caddr_t ptr, int argc, char **argv, int trytags)
{
	struct tcl_arg_info	*item;
	struct tcl_type_info	*type;
	int			i, error;

	/* Try untagged parsing */
	if (trytags) {
		if (argc == 0)
			goto do_tagged;
		if (argv[0][0] == '-') {
			for (i = 0; i < info->nrargs; i++) {
				item = &info->args[i];
				if (!strcmp(item->member, argv[0] + 1))
					goto do_tagged;
			}
		}
	}

	/* Okay, untagged parsing */
	if (info->nrargs != argc)
		return tcl_error(interp, "bad number of arguments");

	for (i = 0; i < info->nrargs; i++) {
		item = &info->args[i];
		type = item->type;

		error = type->parse(interp, type->data,
					ptr + item->offset, argv[i]);

		if (error != TCL_OK)
			return error;
	}

	return TCL_OK;

do_tagged:
	if (argc & 1)
		return tcl_error(interp,
			"odd number of elements in %s args list", info->name);

	while (argc) {
		for (i = 0; i < info->nrargs; i++) {
			item = &info->args[i];
			if (!strcmp(item->member, argv[0] + 1))
				break;
		}
		if (i >= info->nrargs)
			return tcl_error(interp,
				"unknown element %s in %s args list",
				argv[0], info->name);

		type = item->type;
		error = type->parse(interp, type->data,
					ptr + item->offset, argv[1]);
		if (error != TCL_OK)
			return error;

		argc -= 2;
		argv += 2;
	}

	return TCL_OK;
}

void
tcl_struct_init(struct tcl_parm_info *info, caddr_t ptr)
{
	struct tcl_type_info	*type;
	struct tcl_arg_info	*item;
	int			i;

	for (i = 0; i < info->nrargs; i++) {
		item = &info->args[i];
		type = item->type;

		if (type->init)
			type->init(type->data, ptr + item->offset);
	}
}

int
tcl_struct_print(Tcl_Interp *interp, struct tcl_parm_info *info, caddr_t ptr)
{
	struct tcl_type_info	*type;
	struct tcl_arg_info	*item;
	int			i, error;


	for (i = 0; i < info->nrargs; i++) {
		item = &info->args[i];
		type = item->type;

		error = type->print(interp, type->data, ptr + item->offset);
		if (error == TCL_ERROR)
			return error;
		if (error == TCL_BREAK)
			break;
	}

	return TCL_OK;
}

static int
tcl_struct_help(Tcl_Interp *interp, char *proc, struct tcl_parm_info *info)
{
	int	i;

	tcl_return(interp, "usage: %s", proc);
	for (i = 0; i < info->nrargs; i++)
		Tcl_AppendElement(interp, info->args[i].type->name);

	return TCL_ERROR;
}

static int
tcl_type_help(Tcl_Interp *interp, char *proc, struct tcl_type_info *type)
{
	return tcl_error(interp, "usage: %s %s", proc, type->name);
}
