/*
 * shell/async.c
 *
 * Kludge async socket I/O (for RPC services) onto vanilla TCL.
 *
 * The (relatively) clean solution is to arrange for async I/O handling
 * on the socket through SIGIO, and create a Tcl Async handler that gets
 * invoked upon receipt of a signal.
 *
 * The second solution (if async socket I/O is not available) is through
 * polling, which is achieved by tracing command execution and checking
 * for input for every command executed.
 */

#include <signal.h>
#include <nfshell.h>
#include <tcl.h>
#include <shell/rpc.h>
#include <shell/args.h>

#ifdef HAVE_FASYNC
#include <fcntl.h>

static int		fasync = 0;
static Tcl_AsyncHandler	fasyncHandle;

static void		sigasync(int sig);
static int		tcl_async_poll(ClientData, Tcl_Interp *, int);
#else
static int		tracing = 0;
static Tcl_Trace	traceHandle;

static void		tcl_async_poll(ClientData, Tcl_Interp *, int, char *,
				Tcl_CmdProc *, ClientData, int, char **);
#endif

#ifdef HAVE_FASYNC
/*
 * Arrange for async I/O through SIGO
 */
int
tcl_svc_async(Tcl_Interp *interp, int fd, void (*poll)(void))
{
	if (!fasync) {
		struct sigaction	act;

		fasyncHandle = Tcl_AsyncCreate(tcl_async_poll,
						(ClientData) poll);
		act.sa_handler = sigasync;
		act.sa_flags   = SA_RESTART;
		sigemptyset(&act.sa_mask);
		sigaction(SIGIO, &act, NULL);

		fasync = 1;
	}

	if (fcntl(fd, F_SETOWN, getpid())
	 || fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | FASYNC) < 0) {
		return tcl_error(interp, "can't set async I/O on socket");
	}

	return TCL_OK;
}

static void
sigasync(int sig)
{
	Tcl_AsyncMark(fasyncHandle);
}

static int
tcl_async_poll(ClientData cd, Tcl_Interp *interp, int code)
{
	((void (*)(void)) cd)();
	return code;
}
#else
/*
 * Arrange for async socket I/O through command tracing.
 */
int
tcl_svc_async(Tcl_Interp *interp, int fd, void (*poll)(void))
{
	if (!tracing) {
		traceHandle = Tcl_CreateTrace(interp, 1000,
					tcl_async_poll,
					(ClientData) poll);
		tracing = 1;
	}

	return TCL_OK;
}

static void
tcl_async_poll(ClientData cd, Tcl_Interp *interp, int level, char *cmd,
				Tcl_CmdProc *cmdProc, ClientData cmdCd,
				int argc, char **argv)
{
	((void (*)(void)) cd)();
}
#endif
