/*
 * shell/nfs2args.c
 *
 * Define the argument conversion for the NFSv2 binding
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <memory.h>
#include <malloc.h>
#include <nfshell.h>
#include <xdr/nfs.h>
#include <shell/nfs.h>

TCL_DECLARE_TYPE(extern, u32);

/*
 * TCL handling for NFS status
 */
#define tcl_nfs_stat_init	NULL

static int
tcl_nfs_stat_parse(Tcl_Interp *interp, void *d, u32 *s, char *a)
{
	return tcl_error(interp, "NFS: can't parse nfs_stat (not implemented)");
}

static int
tcl_nfs_stat_print(Tcl_Interp *interp, void *dummy, u32 *statp)
{
	char	buffer[20], *errname;

	if (*statp == 0)
		return 0;

	switch (*statp) {
	case NFSERR_PERM:	errname = "NFSERR_PERM"; break;
	case NFSERR_NOENT:	errname = "NFSERR_NOENT"; break;
	case NFSERR_IO:		errname = "NFSERR_IO"; break;
	case NFSERR_NXIO:	errname = "NFSERR_NXIO"; break;
	case NFSERR_ACCES:	errname = "NFSERR_ACCES"; break;
	case NFSERR_EXIST:	errname = "NFSERR_EXIST"; break;
	case NFSERR_NODEV:	errname = "NFSERR_NODEV"; break;
	case NFSERR_NOTDIR:	errname = "NFSERR_NOTDIR"; break;
	case NFSERR_ISDIR:	errname = "NFSERR_ISDIR"; break;
	case NFSERR_FBIG:	errname = "NFSERR_FBIG"; break;
	case NFSERR_NOSPC:	errname = "NFSERR_NOSPC"; break;
	case NFSERR_ROFS:	errname = "NFSERR_ROFS"; break;
	case NFSERR_NAMETOOLONG:errname = "NFSERR_NAMETOOLONG"; break;
	case NFSERR_NOTEMPTY:	errname = "NFSERR_NOTEMPTY"; break;
	case NFSERR_DQUOT:	errname = "NFSERR_DQUOT"; break;
	case NFSERR_STALE:	errname = "NFSERR_STALE"; break;
	case NFSERR_WFLUSH:	errname = "NFSERR_WFLUSH"; break;
	default:
		sprintf(buffer, "nfserr%d", (int) *statp);
		errname = buffer;
		break;
	}

	Tcl_AppendElement(interp, errname);
	return TCL_ERROR;
}

TCL_DEFINE_TYPE(extern, nfs_stat, NULL);

/*
 * TCL handling for nfs_fh
 */
#define tcl_nfs_fh_init		NULL

static int
tcl_nfs_fh_parse(Tcl_Interp *interp, void *d, nfs_fh *fh, char *arg)
{
	unsigned char	byte = 0, c = 0, *data = (unsigned char *) fh;
	int		i;

	memset(fh, 0, sizeof(*fh));

	for (i = 0; arg[i] != '\0' && i < 64; i++) {
		c = arg[i];
		if ('0' <= c && c <= '9')
			c -= '0';
		else if ('a' <= c && c <= 'f')
			c -= 'a' - 10;
		else if ('A' <= c && c <= 'F')
			c -= 'A' - 10;
		else
			return tcl_error(interp,
				"invalid hex character in fh %s", arg);
		if ((i & 1) == 0)
			byte = c;
		else
			*data++ = (byte << 4) | c;
	}

	if (i & 1)
		*data++ = (byte << 4);

	return TCL_OK;
}

static int
tcl_nfs_fh_print(Tcl_Interp *interp, void *dummy, nfs_fh *fh)
{
	char		buffer[65], *sp = buffer;
	unsigned char	*data = (unsigned char *) fh;
	int		i;

	for (i = 0; i < 32; i++, sp += 2)
		sprintf(sp, "%02x", data[i]);
	Tcl_AppendElement(interp, buffer);
	return TCL_OK;
}

/*
 * TCL handling for nfs_ftype
 */
#define tcl_nfs_ftype_init		NULL

static int
tcl_nfs_ftype_parse(Tcl_Interp *interp, void *d, u32 *ftype, char *arg)
{
	if (!strcmp(arg, "REG") || !strcmp(arg, "reg"))
		*ftype = NFREG;
	else
	if (!strcmp(arg, "DIR") || !strcmp(arg, "dir"))
		*ftype = NFDIR;
	else
	if (!strcmp(arg, "BLK") || !strcmp(arg, "blk"))
		*ftype = NFBLK;
	else
	if (!strcmp(arg, "CHR") || !strcmp(arg, "chr"))
		*ftype = NFCHR;
	else
	if (!strcmp(arg, "LNK") || !strcmp(arg, "lnk"))
		*ftype = NFLNK;
	else
	if (!strcmp(arg, "SOCK") || !strcmp(arg, "sock"))
		*ftype = NFSOCK;
	else
	if (!strcmp(arg, "FIFO") || !strcmp(arg, "fifo"))
		*ftype = NFFIFO;
	else
		return tcl_error(interp, "bad NFS ftype %s", arg);
	return TCL_OK;
}

static int
tcl_nfs_ftype_print(Tcl_Interp *interp, void *dummy, u32 *ftype)
{
	char	*name;

	switch (*ftype) {
	case NFNON:	name = "NON";  break;
	case NFREG:	name = "REG";  break;
	case NFDIR:	name = "DIR";  break;
	case NFBLK:	name = "BLK";  break;
	case NFCHR:	name = "CHR";  break;
	case NFLNK:	name = "LNK";  break;
	case NFSOCK:	name = "SOCK"; break;
	case NFFIFO:	name = "FIFO"; break;
	default:
		return tcl_error(interp, "unknown NFS ftype %x", *ftype);
	}
	Tcl_AppendElement(interp, name);
	return TCL_OK;
}

/*
 * Data handling
 */
static void
tcl_nfs_data_init(void *dummy, nfs_data *data)
{
	data->data = NULL;
	data->count = 0;
}

static int
tcl_nfs_data_parse(Tcl_Interp *interp, void *d, nfs_data *data, char *arg)
{
	data->data = (u32 *) arg;
	data->count = strlen(arg);
	return TCL_OK;
}

static int
tcl_nfs_data_print(Tcl_Interp *interp, void *dummy, nfs_data *data)
{
	unsigned char	*ptr = (unsigned char *) data;
	int		i;

	if (data->count > 16 * 1024)
		return tcl_error(interp,
			"king-size data portion (%d bytes)", data->count);
	for (i = 0; i < data->count; i++, ptr++) {
		if ((*ptr & 0x7f) < 32)
			return tcl_error(interp,
				"can't represent data (byte %x found)", *ptr);
	}

	return TCL_OK;
}

/*
 * sattr handling
 */
static void
tcl_nfs_sattr_init(void *dummy, nfs_sattr *attr)
{
	memset(attr, -1, sizeof(*attr));
}

/*
 * Callback handling
 */
struct tcl_nfs_rddata {
	Tcl_DString		dstring;
};

static void
tcl_nfs_callback(Tcl_DString *result, char *name, u32 ino)
{
	char	buffer[512];

	if (name != NULL)
		sprintf(buffer, "%s %ld", name, (unsigned long) ino);
	else
		sprintf(buffer, "EOF");
	Tcl_DStringAppendElement(result, buffer);
}

static void
tcl_nfs_callback_init(void *dummy, nfs_callback *callback)
{
	Tcl_DString	*result;

	result = (Tcl_DString *) malloc(sizeof(*result));
	Tcl_DStringInit(result);

	callback->func = (nfs_dirent_fn_t) tcl_nfs_callback;
	callback->data = result;
}

static int
tcl_nfs_callback_parse(Tcl_Interp *interp, void *d, nfs_callback *c, char *a)
{
	return tcl_error(interp, "can't parse NFS callback spec");
}

static int
tcl_nfs_callback_print(Tcl_Interp *interp, void *dummy, nfs_callback *callback)
{
	Tcl_DString	*result = (Tcl_DString *) callback->data;
	char		*dirs;

	dirs = Tcl_DStringValue(result);
	Tcl_AppendElement(interp, dirs);
	Tcl_DStringFree(result);
	free(result);
	return TCL_OK;
}

#define tcl_nfs_fattr_init		NULL
#define tcl_nfs_attrstat_init		NULL
#define tcl_nfs_sattrargs_init		tcl_struct_init
#define tcl_nfs_diropargs_init		NULL
#define tcl_nfs_diropres_init		NULL
#define tcl_nfs_readlinkres_init	NULL
#define tcl_nfs_readargs_init		NULL
#define tcl_nfs_readres_init		NULL
#define tcl_nfs_writeargs_init		NULL
#define tcl_nfs_createargs_init		tcl_struct_init
#define tcl_nfs_renameargs_init		NULL
#define tcl_nfs_linkargs_init		NULL
#define tcl_nfs_symlinkargs_init	tcl_struct_init
#define tcl_nfs_readdirargs_init	NULL
#define tcl_nfs_readdirres_init		NULL
#define tcl_nfs_fstat_init		NULL
#define tcl_nfs_statfsres_init		NULL

TCL_DEFINE_TYPE(extern, nfs_fh, NULL);
TCL_DEFINE_TYPE(static, nfs_ftype, NULL);
TCL_DEFINE_TYPE(static, nfs_data, NULL);
TCL_DEFINE_TYPE(static, nfs_callback, NULL);

TCL_BEGIN_STRUCT(static, nfs_fattr)
	TCL_MEMBER(nfs_fattr, nfs_ftype,	type),
	TCL_MEMBER(nfs_fattr, u32,		mode),
	TCL_MEMBER(nfs_fattr, u32,		nlink),
	TCL_MEMBER(nfs_fattr, u32,		uid),
	TCL_MEMBER(nfs_fattr, u32,		gid),
	TCL_MEMBER(nfs_fattr, u32,		size),
	TCL_MEMBER(nfs_fattr, u32,		blocksize),
	TCL_MEMBER(nfs_fattr, u32,		rdev),
	TCL_MEMBER(nfs_fattr, u32,		blocks),
	TCL_MEMBER(nfs_fattr, u32,		fsid),
	TCL_MEMBER(nfs_fattr, u32,		fileid),
	TCL_MEMBER(nfs_fattr, timeval,		atime),
	TCL_MEMBER(nfs_fattr, timeval,		mtime),
	TCL_MEMBER(nfs_fattr, timeval,		ctime),
TCL_END_STRUCT(static, nfs_fattr);
TCL_DEFINE_STRUCT(static, nfs_fattr);

TCL_BEGIN_STRUCT(static, nfs_sattr)
	TCL_MEMBER(nfs_sattr, u32,		mode),
	TCL_MEMBER(nfs_sattr, u32,		uid),
	TCL_MEMBER(nfs_sattr, u32,		gid),
	TCL_MEMBER(nfs_sattr, u32,		size),
	TCL_MEMBER(nfs_sattr, timeval,		atime),
	TCL_MEMBER(nfs_sattr, timeval,		mtime),
TCL_END_STRUCT(static, nfs_sattr)
TCL_DEFINE_STRUCT(static, nfs_sattr);

TCL_BEGIN_STRUCT(static, nfs_attrstat)
	TCL_MEMBER(nfs_attrstat, nfs_stat,	nfsstat),
	TCL_MEMBER(nfs_attrstat, nfs_fattr,	fattr),
TCL_END_STRUCT(static, nfs_attrstat);
TCL_DEFINE_STRUCT(extern, nfs_attrstat);

TCL_BEGIN_STRUCT(static, nfs_sattrargs)
	TCL_MEMBER(nfs_sattrargs, nfs_fh,	fh),
	TCL_MEMBER(nfs_sattrargs, nfs_sattr,	sattr),
TCL_END_STRUCT(static, nfs_sattrargs);
TCL_DEFINE_STRUCT(extern, nfs_sattrargs);

TCL_BEGIN_STRUCT(static, nfs_diropargs)
	TCL_MEMBER(nfs_diropargs, nfs_fh,	fh),
	TCL_MEMBER(nfs_diropargs, string,	name),
TCL_END_STRUCT(static, nfs_diropargs);
TCL_DEFINE_STRUCT(extern, nfs_diropargs);

TCL_BEGIN_STRUCT(static, nfs_diropres)
	TCL_MEMBER(nfs_diropres, nfs_stat,	nfsstat),
	TCL_MEMBER(nfs_diropres, nfs_fh,	fh),
	TCL_MEMBER(nfs_diropres, nfs_fattr,	fattr),
TCL_END_STRUCT(static, nfs_diropres);
TCL_DEFINE_STRUCT(extern, nfs_diropres);

TCL_BEGIN_STRUCT(static, nfs_readlinkres)
	TCL_MEMBER(nfs_readlinkres, nfs_stat,	nfsstat),
	TCL_MEMBER(nfs_readlinkres, string,	path),
TCL_END_STRUCT(static, nfs_readlinkres);
TCL_DEFINE_STRUCT(extern, nfs_readlinkres);

TCL_BEGIN_STRUCT(static, nfs_readargs)
	TCL_MEMBER(nfs_readargs, nfs_fh,	fh),
	TCL_MEMBER(nfs_readargs, u32,		offset),
	TCL_MEMBER(nfs_readargs, u32,		count),
TCL_END_STRUCT(static, nfs_readargs);
TCL_DEFINE_STRUCT(extern, nfs_readargs);

TCL_BEGIN_STRUCT(static, nfs_readres)
	TCL_MEMBER(nfs_readres, nfs_stat,	nfsstat),
	TCL_MEMBER(nfs_readres, nfs_fattr,	fattr),
	TCL_MEMBER(nfs_readres, nfs_data,	data),
TCL_END_STRUCT(static, nfs_readres);
TCL_DEFINE_STRUCT(extern, nfs_readres);

TCL_BEGIN_STRUCT(static, nfs_writeargs)
	TCL_MEMBER(nfs_writeargs, nfs_fh,	fh),
	TCL_MEMBER(nfs_writeargs, u32,		offset),
	TCL_MEMBER(nfs_writeargs, nfs_data,	data),
TCL_END_STRUCT(static, nfs_writeargs);
TCL_DEFINE_STRUCT(extern, nfs_writeargs);

TCL_BEGIN_STRUCT(static, nfs_createargs)
	TCL_MEMBER(nfs_createargs, nfs_fh,	fh),
	TCL_MEMBER(nfs_createargs, string,	name),
	TCL_MEMBER(nfs_createargs, nfs_sattr,	sattr),
TCL_END_STRUCT(static, nfs_createargs);
TCL_DEFINE_STRUCT(extern, nfs_createargs);

TCL_BEGIN_STRUCT(static, nfs_renameargs)
	TCL_MEMBER(nfs_renameargs, nfs_fh,	ffh),
	TCL_MEMBER(nfs_renameargs, string,	fname),
	TCL_MEMBER(nfs_renameargs, nfs_fh,	tfh),
	TCL_MEMBER(nfs_renameargs, string,	tname),
TCL_END_STRUCT(static, nfs_renameargs);
TCL_DEFINE_STRUCT(extern, nfs_renameargs);

TCL_BEGIN_STRUCT(static, nfs_linkargs)
	TCL_MEMBER(nfs_linkargs, nfs_fh,	ffh),
	TCL_MEMBER(nfs_linkargs, nfs_fh,	tfh),
	TCL_MEMBER(nfs_linkargs, string,	tname),
TCL_END_STRUCT(static, nfs_linkargs);
TCL_DEFINE_STRUCT(extern, nfs_linkargs);

TCL_BEGIN_STRUCT(static, nfs_symlinkargs)
	TCL_MEMBER(nfs_symlinkargs, nfs_fh,	ffh),
	TCL_MEMBER(nfs_symlinkargs, string,	fname),
	TCL_MEMBER(nfs_symlinkargs, string,	path),
	TCL_MEMBER(nfs_symlinkargs, nfs_sattr,	sattr),
TCL_END_STRUCT(static, nfs_symlinkargs);
TCL_DEFINE_STRUCT(extern, nfs_symlinkargs);

TCL_BEGIN_STRUCT(static, nfs_readdirargs)
	TCL_MEMBER(nfs_readdirargs, nfs_fh,	fh),
	TCL_MEMBER(nfs_readdirargs, u32,	cookie),
	TCL_MEMBER(nfs_readdirargs, u32,	count),
TCL_END_STRUCT(static, nfs_readdirargs);
TCL_DEFINE_STRUCT(extern, nfs_readdirargs);

TCL_BEGIN_STRUCT(static, nfs_readdirres)
	TCL_MEMBER(nfs_readdirres, nfs_stat,	nfsstat),
	TCL_MEMBER(nfs_readdirres, nfs_callback,iter),
TCL_END_STRUCT(static, nfs_readdirres);
TCL_DEFINE_STRUCT(extern, nfs_readdirres);

TCL_BEGIN_STRUCT(static, nfs_fstat)
	TCL_MEMBER(nfs_fstat, u32,		tsize),
	TCL_MEMBER(nfs_fstat, u32,		bsize),
	TCL_MEMBER(nfs_fstat, u32,		blocks),
	TCL_MEMBER(nfs_fstat, u32,		bfree),
	TCL_MEMBER(nfs_fstat, u32,		bavail),
TCL_END_STRUCT(static, nfs_fstat);
TCL_DEFINE_STRUCT(static, nfs_fstat);

TCL_BEGIN_STRUCT(static, nfs_statfsres)
	TCL_MEMBER(nfs_statfsres, nfs_stat,	nfsstat),
	TCL_MEMBER(nfs_statfsres, nfs_fstat,	stats),
TCL_END_STRUCT(static, nfs_statfsres);
TCL_DEFINE_STRUCT(extern, nfs_statfsres);
