/*
 * shell/rpc.c
 *
 * Generic RPC handling
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <malloc.h>
#include <memory.h>
#include <nfshell.h>
#include <xdr/portmap.h>
#include <xdr/mount.h>
#include <xdr/nfs.h>
#include <shell/args.h>
#include <shell/portmap.h>
#include <shell/mount.h>
#include <shell/nfs.h>
#include <shell/nlm.h>
#include <shell/rpc.h>

static int	tcl_rpc_switch(ClientData, Tcl_Interp *, int, char **);
static int	tcl_rpc_create(Tcl_Interp *, int, char **);
static int	tcl_svc_create(Tcl_Interp *, int, char **);
static int	tcl_svc_switch(ClientData, Tcl_Interp *, int, char **);
static int	tcl_gethostbyname(char *, sockaddr_in *);
static int	tcl_getrpcbyname(char *, u32 *);
static int	tcl_pmap_getport(Tcl_Interp *, sockaddr_in *, u32, u32, u32 *);

/*
 * Initialize the RPC tcl binding
 */
void
tcl_rpc_register(Tcl_Interp *interp)
{
	Tcl_CreateCommand(interp, "rpc", tcl_rpc_switch, NULL, NULL);
}

/*
 * Dispatch calls to the RPC handlers
 */
static int
tcl_rpc_switch(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
	if (argc < 2)
		return tcl_error(interp, "rpc: minor command missing");

	if (!strcmp(argv[1], "clnt_create") || !strcmp(argv[1], "create"))
		return tcl_rpc_create(interp, argc - 2, argv + 2);

	if (!strcmp(argv[1], "svc_create"))
		return tcl_svc_create(interp, argc - 2, argv + 2);

	if (!strcmp(argv[1], "setuid")) {
		int	uid;

		if (argc != 3)
			return tcl_error(interp, "rpc setuid: bad # of args");
		if (Tcl_GetInt(interp, argv[2], &uid) != TCL_OK)
			return tcl_error(interp,
				"rpc setuid: bad uid argument %s", argv[2]);
		return tcl_return(interp, "%d", rpc_setuid(uid));
	}

	if (!strcmp(argv[1], "setgid")) {
		int	gid;

		if (argc != 3)
			return tcl_error(interp, "rpc setgid: bad # of args");
		if (Tcl_GetInt(interp, argv[2], &gid) != TCL_OK)
			return tcl_error(interp,
				"rpc setgid: bad gid argument %s", argv[2]);
		return tcl_return(interp, "%d", rpc_setgid(gid));
	}

#if 1
	return tcl_error(interp, "rpc: unknown minor command %s", argv[1]);
#else
	Tcl_SetResult(interp, "unknown minor", TCL_VOLATILE);
	return TCL_ERROR;
#endif
}


/*
 * tcl_rpc_create
 *
 * rpc create host prog [vers [port]]
 */
static int
tcl_rpc_create(Tcl_Interp *interp, int argc, char **argv)
{
	sockaddr_in	sin;
	struct rpcprog	*program;
	u32		prog, vers, port = 0;
	int		ivers = 0, iport = 0;
	void		(*regfunc)(Tcl_Interp *, struct rpcprog *);

	if (argc < 2 || argc > 4)
		return tcl_error(interp, "bad number of arguments");

	if (tcl_gethostbyname(argv[0], &sin) == 0)
		return tcl_error(interp, "unknown host %s", argv[0]);

	if (tcl_getrpcbyname(argv[1], &prog) == 0)
		return tcl_error(interp, "unknown RPC service %s", argv[1]);

	if (argc > 2 && Tcl_GetInt(interp, argv[2], &ivers) != TCL_OK)
		return tcl_error(interp, "bad version number %s", argv[2]);
	vers = (u32) ivers;

	if (argc > 3 && Tcl_GetInt(interp, argv[3], &iport) != TCL_OK)
		return tcl_error(interp, "bad port number %s", argv[2]);

	if (!port
	 && tcl_pmap_getport(interp, &sin, prog, vers, &port) != TCL_OK)
		return TCL_ERROR;

	if (vers == 0) {
		/* should probe for available versions here... 
		 * For now we just guess and fail if we're wrong */
		vers = 1;
	}

	switch (prog) {
	case NFS_PROGRAM:
		regfunc = tcl_nfs_register;
		break;
	case MOUNT_PROGRAM:
		regfunc = tcl_mnt_register;
		break;
	case PMAP_PROGRAM:
		regfunc = tcl_pmap_register;
		break;
	case NLM_PROGRAM:
		regfunc = tcl_nlm_register;
		break;
	default:
		return tcl_error(interp,
			"no tcl binding for RPC service %s", argv[1]);
	}

	if (!(program = rpc_clntcreate(&sin, prog, vers, port)))
		return tcl_error(interp,
			"no protocol for RPC service %s", argv[1]);

	regfunc(interp, program);
	return TCL_OK;
}

/*
 * Create an RPC service.
 *
 * rpc svc_create prog vers
 */
static int
tcl_svc_create(Tcl_Interp *interp, int argc, char **argv)
{
	static int	svc_cmdnum = 1;
	struct rpcserv	*serv;
	char		buffer[20];
	u32		prog, vers;
	int		ivers;

	if (argc != 2)
		return tcl_error(interp, "bad number of arguments");

	if (tcl_getrpcbyname(argv[0], &prog) == 0)
		return tcl_error(interp, "unknown RPC service %s", argv[0]);

	if (Tcl_GetInt(interp, argv[1], &ivers) != TCL_OK
	 || ivers < 0 || ivers > 10)
		return tcl_error(interp, "bad version number %s", argv[1]);
	vers = (u32) ivers;

	if (!(serv = rpc_svccreate(prog, vers)))
		return tcl_error(interp,
				"no protocol for RPC service %s", argv[0]);

	if (tcl_svc_async(interp, serv->sv_sock, biod_svc_check) != TCL_OK) {
		rpc_svcdestroy(serv);
		return TCL_ERROR;
	}

	sprintf(buffer, "%s%d", serv->sv_name, svc_cmdnum++);
	Tcl_CreateCommand(interp, buffer, tcl_svc_switch,
				(ClientData) serv,
				(Tcl_CmdDeleteProc *) rpc_svcdestroy);

	return tcl_return(interp, buffer);
}

static int
tcl_svc_switch(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
	struct rpcserv	*serv = (struct rpcserv *) cd;

	if (argc < 2)
		return tcl_error(interp,
				"%s: minor command required", argv[0]);

	if (!strcmp(argv[1], "port")) {
		if (argc != 2)
			return tcl_error(interp, "%s %s: too many args",
					argv[0], argv[1]);
		return tcl_return(interp, "%d", serv->sv_port);
	}

	return tcl_error(interp, "%s: unknown method %s", argv[0], argv[1]);
}

static int
tcl_gethostbyname(char *hname, sockaddr_in *sin)
{
	struct hostent	*hp;

	if (!(hp = gethostbyname(hname)))
		return 0;
	if (hp->h_addrtype != AF_INET || hp->h_length != 4)
		return 0;

	memset(sin, 0, sizeof(*sin));
	sin->sin_family = AF_INET;
	sin->sin_addr = *(struct in_addr *) hp->h_addr;
	return 1;
}

static int
tcl_getrpcbyname(char *pname, u32 *progp)
{
	struct rpcent	*rp;

	if ('0' <= *pname && *pname <= '9') {
		*progp = atoi(pname);
		return 1;
	}
	if (!(rp = getrpcbyname(pname)))
		return 0;
	*progp = rp->r_number;
	return 1;
}

static int
tcl_pmap_getport(Tcl_Interp *interp, sockaddr_in *sin, u32 prog, u32 vers, u32 *portp)
{
	struct rpcprog	*pmapper;
	pmap_map	pmap;
	u32		rpcstat;
	int		status;

	pmapper = rpc_clntcreate(sin, PMAP_PROGRAM, PMAP_VERSION, PMAP_PORT);
	if (pmapper == NULL)
		return tcl_error(interp, "can't create portmapper handle");

	pmap.pm_prog = prog;
	pmap.pm_vers = vers;
	pmap.pm_prot = IPPROTO_UDP;
	pmap.pm_port = 0;

	rpcstat = rpc_call(pmapper, PMAPPROC_GETPORT, &pmap, portp);
	if (rpcstat != RPC_SUCCESS) {
		status = tcl_error(interp,
			"couldn't contact portmapper on %s: %s",
			inet_ntoa(sin->sin_addr), rpc_strerror(rpcstat));
	} else {
		status = TCL_OK;
	}

	rpc_clntdestroy(pmapper);

	return status;
}
