/*
 * shell/svc.c
 *
 * RPC service support from TCL. This is a big kludge since vanilla TCL
 * doesn't have a top-level loop that checks for things like input events.
 * We therefore arrange for TCL command tracing as soon as a server is
 * instantiated. The trace procedure checks to see whether there are
 * any RPC requests pending on the service port, and if so, evaluates
 * them.
 *
 * Set FASYNC for socket, install signal handler, and create Tcl async
 * handler that gets invoked upon receipt of signal.
 */

#include <signal.h>
#include <nfshell.h>
#include <tcl.h>
#include <shell/args.h>

#ifdef HAVE_FASYNC
#include <fcntl.h>

static int		fasync = 0;
static Tcl_AsyncHandler	fasyncHandle;

static void		sigasync(int sig);
static int		tcl_rpc_svcpoll(ClientData, Tcl_Interp *, int);
#else
static int		tracing = 0;
static Tcl_Trace	traceHandle;
#endif

/*
 * Create an RPC service.
 *
 * rpc svc_create prog vers
 */
static int
tcl_svc_create(Tcl_Interp *interp, int argc, char **argv)
{
	struct rpcsvc	*serv;
	u32		prog, vers;
	int		ivers;
#ifdef HAVE_FASYNC
	int		flags;
#endif

	if (argc != 2)
		return tcl_error(interp, "bad number of arguments");

	if (tcl_getrpcbyname(argv[0], &prog) == 0)
		return tcl_error(interp, "unknown RPC service %s", argv[0]);

	if (Tcl_GetInt(interp, argv[1], &ivers) != TCL_OK
	 || ivers < 0 || ivers > 10)
		return tcl_error(interp, "bad version number %s", argv[1]);
	vers = (u32) ivers;

	if (!(serv = rpc_svccreate(prog, vers, interp)))
		return tcl_error(interp,
				"no protocol for RPC service %s", argv[0]);

#ifdef HAVE_FASYNC
	if (!fasync) {
		struct sigaction	act;

		fasyncHandle = Tcl_AsyncCreate(tcl_rpc_svcpoll, NULL);
		fasync = 1;
		act.sa_handler = sigasync;
		act.sa_flags   = SA_RESTART;
		sigemptyset(&act.sa_mask);
		sigaction(SIGIO, &act, NULL);
	}

	if (fcntl(serv->sv_sock, F_GETFL, &flags) < 0
	 || fcntl(serv->sv_sock, F_SETFL, flags | FASYNC) < 0) {
		rpc_svcdestroy(serv);
		return tcl_error(interp,
				"unable to set FASYNC on server socket!");
	}
#else
	if (!tracing) {
		traceHandle = Tcl_CreateTrace(interp, 1000,
					tcl_rpc_svccheck, NULL);
	}
#endif

	return TCL_OK;
}

#ifdef HAVE_FASYNC
static void
sigasync(int sig)
{
	Tcl_AsyncMark(fasyncHandle);
}

static int
tcl_rpc_svcpoll(ClientData cd, Tcl_Interp *interp, int code)
{
	biod_svc_check();
	return code;
}
#else
static void
tcl_rpc_svccheck(ClientData cd, Tcl_Interp *interp, int level, char *cmd,
				Tcl_CmdProc *cmdProc, ClientData cmdCd,
				int argc, char **argv)
{
	biod_svc_check();
}
#endif
