##################################################################
#
# tcl/tests.tcl
#
# Support functions for tcl test suite
#
# Copyright (C) 1996, 1997 Olaf Kirch <okir@monad.swb.de>
#
##################################################################

proc FILE {file args} {

	global	slave_clnts

	if [interp exists test] {
		interp delete test
	}

	set slave_clnts {}

	if [catch {
		interp create -- test
		interp alias test rpc    {} slave_rpc test
		interp share test stdout {}
		interp share test stderr {}
		interp eval  test source tcl/tests.tcl
		interp eval  test RUN    $file $args
	} error] {
		puts "$error";
	}

	interp delete test

	# Delete all clients created on behalf of the slave interp
	foreach clnt $slave_clnts {
		$clnt delete
	}
}

##################################################################
# Create an RPC service and export it to the slave interp
# This procedure is executed in the parent interp
##################################################################
proc slave_rpc {slave args} {

	global	slave_clnts

	set clnt [eval rpc $args]
	interp alias test $clnt {} $clnt

	if {[lindex $args 0] == "create"} {
		lappend slave_clnts $clnt
	}

	return $clnt
}

##################################################################
# Run the test
##################################################################
proc RUN {file args} {

	global	ARGS verbose

	set newargs {}
	set verbose 0
	foreach arg $args {
		switch -glob -- $arg {
		--verbose	{ set verbose 1; }
		*		{ lappend newargs $arg; }
		}
	}

	set ARGS $newargs;
	source $file
}

##################################################################
# set test arguments
##################################################################
proc ARGS {args} {

	puts "ARGS $args"
	if [llength $args] {
		uplevel #0 set ARGS $args
	}
}

##################################################################
# Initialize a test and parse arguments
##################################################################
proc INIT {body} {

	uplevel #0 eval $body
}

##################################################################
# Start a new test
##################################################################
proc TEST {ident args} {

	global	cur_test cur_title cur_skip cur_cond

	set cur_test	$ident
	set cur_title	$args
	set cur_skip	0
	set cur_cond    1

	global	$cur_test
	set $cur_test	0

	proc __succ__ {} {}
	proc __fail__ {} {}
}

##################################################################
# Evaluate a precondition
##################################################################
proc COND {args} {

	global	cur_cond

	set expr ""
	foreach token $args {
		switch -glob $token {
		"[a-z]*"	{ set token "\$$token"; }
		}
		append expr " $token"
	}
	set cur_cond [uplevel #0 eval expr $expr]
}

##################################################################
##################################################################
proc BODY {body} {

	global	cur_test cur_cond cur_title cur_printed
	global	$cur_test

	set cur_printed 0

	if {$cur_cond == 0} {
		NOTICE "precondition false, skipping."
		return
	}

	set $cur_test 1

	set failed [catch {
		set result [uplevel #0 if {1} [list $body]]
	} errinfo]
	if {$failed} {
		if {$errinfo == "" && $cur_title == ""} {
			__print_message__ "Test failed."
		} elseif {$errinfo != ""} {
			__print_message__ "$errinfo"
		}
		set $cur_test 0
		__fail__
	} else {
		__succ__
		set $cur_test 1
	}
}

##################################################################
##################################################################
proc SUCC {body} {

	proc __succ__ {} $body
}

##################################################################
##################################################################
proc FAIL {body} {

	proc __fail__ {} $body
}

proc WARNFAIL {msg} {

	proc __fail__ {} [list WARN $msg]
}

##################################################################
##################################################################
proc ABORT {} {

	global	cur_test

	return -code error "Test failed: ABORT called in $cur_test"
}

##################################################################
##################################################################
proc ERROR {msg} {

	__print_message__ $msg
	return -code error ""
}

##################################################################
##################################################################
proc WARN {msg} {

	__print_message__ $msg
}

##################################################################
##################################################################
proc NOTICE {msg} {

	global verbose

	if {$verbose} {
		__print_message__ $msg
	}
}

##################################################################
##################################################################
proc __print_message__ {msg} {

	global	cur_printed cur_title

	if {!$cur_printed && $cur_title != ""} {
		__print_line__ $cur_title
	}
	__print_line__ $msg
}

proc __print_line__ {msg} {

	global	cur_printed cur_test

	if {!$cur_printed} {
		puts "[format %-10s $cur_test:] $msg"
	} else {
		puts "           $msg"
	}

	set cur_printed 1
}
