/*
 * xdr/nfs.c
 *
 * NFSv2 XDR functions.
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include "nfshell.h"
#include "macros.h"
#include "xdr/nfs.h"

#define ENC	1
#define DEC	0

int	nfsstat;

/*
 * XDR encode/decode functions.
 */
static u32 *
xdr_nfs_fhandle(u32 *msg, int enc, nfs_fh *fh)
{
	if (enc)
		memcpy(msg, fh, sizeof(*fh));
	else
		memcpy(fh, msg, sizeof(*fh));
#if 0
	if (enc) {
		int	i;

		printf("FH   ");
		for (i = 0; i < 32; i++)
			printf("%02x", (unsigned char) fh->data[i]);
		printf("\n");
	}
#endif
	msg += (sizeof(*fh) + 3) >> 2;
	return msg;
}

static u32 *
xdr_nfs_sattr(u32 *msg, int enc, nfs_sattr *sattr)
{
	*msg++ = htonl(sattr->mode);
	*msg++ = htonl(sattr->uid);
	*msg++ = htonl(sattr->gid);
	*msg++ = htonl(sattr->size);
	*msg++ = htonl(sattr->atime.tv_sec);
	*msg++ = htonl(sattr->atime.tv_usec);
	*msg++ = htonl(sattr->mtime.tv_sec);
	*msg++ = htonl(sattr->mtime.tv_usec);
	return msg;
}

static u32 *
xdr_nfs_fname(u32 *msg, int enc, char *fname)
{
	int	len;

	if (enc) {
		len = strlen(fname);
		*msg++ = htonl(len);
		strcpy((char *) msg, fname);
	} else {
		len = ntohl(*msg++);
		memmove(fname, msg, len);
		fname[len] = 0;
	}
	msg += ((len + 3) >> 2);
	return msg;
}

static u32 *
xdr_nfs_fattr(u32 *msg, int enc, nfs_fattr *fattr)
{
	fattr->type = ntohl(*msg++);
	fattr->mode = ntohl(*msg++);
	fattr->nlink = ntohl(*msg++);
	fattr->uid = ntohl(*msg++);
	fattr->gid = ntohl(*msg++);
	fattr->size = ntohl(*msg++);
	fattr->blocksize = ntohl(*msg++);
	fattr->rdev = ntohl(*msg++);
	fattr->blocks = ntohl(*msg++);
	fattr->fsid = ntohl(*msg++);
	fattr->fileid = ntohl(*msg++);
	fattr->atime.tv_sec = ntohl(*msg++);
	fattr->atime.tv_usec = ntohl(*msg++);
	fattr->mtime.tv_sec = ntohl(*msg++);
	fattr->mtime.tv_usec = ntohl(*msg++);
	fattr->ctime.tv_sec = ntohl(*msg++);
	fattr->ctime.tv_usec = ntohl(*msg++);
	return msg;
}

static u32 *
xdr_nfs_fstat(u32 *msg, int enc, nfs_fstat *fstat)
{
	fstat->tsize = ntohl(*msg++);
	fstat->bsize = ntohl(*msg++);
	fstat->blocks = ntohl(*msg++);
	fstat->bfree = ntohl(*msg++);
	fstat->bavail = ntohl(*msg++);
	return msg;
}

static inline unsigned int
xdr_nfs_reply(struct rpcreq *req, u32 **msgp, u32 *statp)
{
	return ((*statp = ntohl(*((*msgp)++))) == NFS_OK);
}

/*
 * NFS async procedures.
 */
static u32 *
xdr_nfs_null(struct rpcreq *req, u32 *msg, void *dummy)
{
	return msg;
}

static u32 *
xdr_nfs_nullres(struct rpcreq *req, u32 *msg, void *dummy)
{
	return msg;
}

static u32 *
xdr_nfs_fh(struct rpcreq *req, u32 *msg, nfs_fh *argp)
{
	return xdr_nfs_fhandle(msg, ENC, argp);
}

static u32 *
xdr_nfs_attrstat(struct rpcreq *req, u32 *msg, nfs_attrstat *resp)
{
	if ((xdr_nfs_reply(req, &msg, &resp->nfsstat)) != 0) {
		msg = xdr_nfs_fattr(msg, DEC, &resp->fattr);
	}
	return msg;
}

static u32 *
xdr_nfs_sattrargs(struct rpcreq *req, u32 *msg, nfs_sattrargs *argp)
{
	msg = xdr_nfs_fhandle(msg, ENC, &argp->fh);
	msg = xdr_nfs_sattr(msg, ENC, &argp->sattr);
	return msg;
}

static u32 *
xdr_nfs_stat(struct rpcreq *req, u32 *msg, u32 *statp)
{
	xdr_nfs_reply(req, &msg, statp);
	return msg;
}

static u32 *
xdr_nfs_diropargs(struct rpcreq *req, u32 *msg, nfs_diropargs *argp)
{
	msg = xdr_nfs_fhandle(msg, ENC, &argp->fh);
	msg = xdr_nfs_fname(msg, ENC, argp->name);
	return msg;
}

static u32 *
xdr_nfs_diropres(struct rpcreq *req, u32 *msg, nfs_diropres *resp)
{
	if (xdr_nfs_reply(req, &msg, &resp->nfsstat) != 0) {
		msg = xdr_nfs_fhandle(msg, DEC, &resp->fh);
		msg = xdr_nfs_fattr(msg, DEC, &resp->fattr);
	}
	return msg;
}

static u32 *
xdr_nfs_readlinkres(struct rpcreq *req, u32 *msg, nfs_readlinkres *resp)
{
	if (xdr_nfs_reply(req, &msg, &resp->nfsstat) != 0) {
		resp->path = (char *) msg;
		msg = xdr_nfs_fname(msg, DEC, resp->path);
	}
	return msg;
}

static u32 *
xdr_nfs_readargs(struct rpcreq *req, u32 *msg, nfs_readargs *argp)
{
	msg = xdr_nfs_fhandle(msg, ENC, &argp->fh);
	*msg++ = htonl(argp->offset);
	*msg++ = htonl(argp->count);
	*msg++ = htonl(argp->count);
	return msg;
}

static u32 *
xdr_nfs_readres(struct rpcreq *req, u32 *msg, nfs_readres *resp)
{
	nfs_data	*data = &resp->data;

	if (xdr_nfs_reply(req, &msg, &resp->nfsstat) != 0) {
		msg = xdr_nfs_fattr(msg, DEC, &resp->fattr);
		data->count = ntohl(*msg++);
		if (data->data)
			memcpy(data->data, msg, data->count);
		msg += (data->count + 3) >> 2;
	}
	return msg;
}

static u32 *
xdr_nfs_writeargs(struct rpcreq *req, u32 *msg, nfs_writeargs *argp)
{
	nfs_data	*data = &argp->data;

	msg = xdr_nfs_fhandle(msg, ENC, &argp->fh);
	*msg++ = htonl(argp->offset);
	*msg++ = htonl(argp->offset);
	*msg++ = htonl(data->count);
	*msg++ = htonl(data->count);	/* opqaue data count */
	if (data->data) {
		memcpy(msg, data->data, data->count);
	} else {
		memset(msg, 0, data->count);
	}
	msg += (data->count + 3) >> 2;
	return msg;
}

static u32 *
xdr_nfs_createargs(struct rpcreq *req, u32 *msg, nfs_createargs *argp)
{
	msg = xdr_nfs_fhandle(msg, ENC, &argp->fh);
	msg = xdr_nfs_fname(msg, ENC, argp->name);
	msg = xdr_nfs_sattr(msg, ENC, &argp->sattr);
	return msg;
}

static u32 *
xdr_nfs_renameargs(struct rpcreq *req, u32 *msg, nfs_renameargs *argp)
{
	msg = xdr_nfs_fhandle(msg, ENC, &argp->ffh);
	msg = xdr_nfs_fname(msg, ENC, argp->fname);
	msg = xdr_nfs_fhandle(msg, ENC, &argp->tfh);
	msg = xdr_nfs_fname(msg, ENC, argp->tname);
	return msg;
}

static u32 *
xdr_nfs_linkargs(struct rpcreq *req, u32 *msg, nfs_linkargs *argp)
{
	msg = xdr_nfs_fhandle(msg, ENC, &argp->ffh);
	msg = xdr_nfs_fhandle(msg, ENC, &argp->tfh);
	msg = xdr_nfs_fname(msg, ENC, argp->tname);
	return msg;
}

/* man, this is boring */
static u32 *
xdr_nfs_symlinkargs(struct rpcreq *req, u32 *msg, nfs_symlinkargs *argp)
{
	msg = xdr_nfs_fhandle(msg, ENC, &argp->ffh);
	msg = xdr_nfs_fname(msg, ENC, argp->fname);
	msg = xdr_nfs_fname(msg, ENC, argp->path);
	msg = xdr_nfs_sattr(msg, ENC, &argp->sattr);
	return msg;
}

static u32 *
xdr_nfs_readdirargs(struct rpcreq *req, u32 *msg, nfs_readdirargs *argp)
{
	msg = xdr_nfs_fhandle(msg, ENC, &argp->fh);
	*msg++ = htonl(argp->cookie);
	*msg++ = htonl(argp->count);
	return msg;
}

static u32 *
xdr_nfs_readdirres(struct rpcreq *req, u32 *msg, nfs_readdirres *resp)
{
	nfs_callback	*iter = &resp->iter;
	char		*name;
	u32		cookie;
	int		len;

	if (xdr_nfs_reply(req, &msg, &resp->nfsstat) == 0)
		return msg;

	while (*msg++) {
		msg++;			/* fileid */
		len = ntohl(*msg++);	/* name length */
		name = (char *) msg;
		msg += (len + 3) >> 2;
		cookie = ntohl(*msg++);	/* readdir cookie */
		name[len] = '\0';
		iter->func(iter->data, name, cookie);
	}
	if (ntohl(*msg++))
		iter->func(iter->data, NULL, 0);

	return msg;
}

static u32 *
xdr_nfs_statfsres(struct rpcreq *req, u32 *msg, nfs_statfsres *resp)
{
	if (xdr_nfs_stat(req, msg, &resp->nfsstat) != 0) {
		msg = xdr_nfs_fstat(msg, DEC, &resp->stats);
	}
	return msg;
}

#define RPC_PROTOCOL		NFS_PROTOCOL
#define RPC_PROT_NAME		nfs
#include <xdr/nfs_prot.h>
#include <xdr/xdrdefs.h>
