/*
 * xdr/nlm.c
 *
 * XDR functions for the NLM protocol
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include "nfshell.h"
#include "macros.h"
#include "xdr/nlm.h"
#include "xdr/nlm_prot.h"

#define ENC	1
#define DEC	0

static u32 *
xdr_encode_string(u32 *msg, char *str)
{
	int	len;

	*msg++ = htonl(len = strlen(str));
	memcpy(msg, str, len);
	return msg + ((len + 3) >> 2);
}

static u32 *
xdr_decode_string(u32 *msg, char **str)
{
	int	len;

	len = ntohl(*msg++);
	if ((len & 3) == 0)
		memmove(msg-1, msg, len);
	*str = (char *) msg;
	(*str)[len] = '\0';

	return msg + ((len + 3) >> 2);
}

static u32 *
xdr_encode_fhandle(u32 *msg, struct nfs_fh *fh)
{
	*msg++ = htonl(32);
	memcpy(msg, fh, 32);
	return msg + 8;
}

static u32 *
xdr_decode_fhandle(u32 *msg, struct nfs_fh *fh)
{
	int	len;

	len = ntohl(*msg++);
	memcpy(fh, msg, 32);
	return msg + ((len + 3) >> 2);
}

static u32 *
xdr_encode_nlm_lock(u32 *msg, nlm_lock *lock)
{
	msg = xdr_encode_string(msg, lock->caller);
	msg = xdr_encode_fhandle(msg, &lock->fhandle);
	msg = xdr_encode_string(msg, lock->owner);
	*msg++ = htonl(lock->pid);
	*msg++ = htonl(lock->offset);
	*msg++ = htonl(lock->length);
	return msg;
}

static u32 *
xdr_decode_nlm_lock(u32 *msg, nlm_lock *lock)
{
	msg = xdr_decode_string(msg, &lock->caller);
	msg = xdr_decode_fhandle(msg, &lock->fhandle);
	msg = xdr_decode_string(msg, &lock->owner);
	lock->pid = ntohl(*msg++);
	lock->offset = ntohl(*msg++);
	lock->length = ntohl(*msg++);
	return msg;
}

static u32 *
xdr_encode_nlm_holder(u32 *msg, nlm_holder *hold)
{
	*msg++ = htonl(hold->excl);
	*msg++ = htonl(hold->pid);
	msg = xdr_encode_string(msg, hold->owner);
	*msg++ = htonl(hold->offset);
	*msg++ = htonl(hold->length);
	return msg;
}

static u32 *
xdr_decode_nlm_holder(u32 *msg, nlm_holder *hold)
{
	hold->excl = ntohl(*msg++);
	hold->pid = ntohl(*msg++);
	msg = xdr_decode_string(msg, &hold->owner);
	hold->offset = ntohl(*msg++);
	hold->length = ntohl(*msg++);
	return msg;
}

/*
 * NLM procedures
 */
static u32 *
xdr_nlm_testargs(struct rpcreq *req, u32 *msg, nlm_testargs *argp)
{
	*msg++ = htonl(argp->cookie);
	*msg++ = htonl(argp->excl);
	return xdr_encode_nlm_lock(msg, &argp->lock);
}

static u32 *
xdrsvc_nlm_testargs(struct svcreq *req, u32 *msg, nlm_testargs *argp)
{
	argp->cookie = ntohl(*msg++);
	argp->excl = ntohl(*msg++);
	return xdr_decode_nlm_lock(msg, &argp->lock);
}

static u32 *
xdr_nlm_lockargs(struct rpcreq *req, u32 *msg, nlm_lockargs *argp)
{
	*msg++ = htonl(argp->cookie);
	*msg++ = htonl(argp->block);
	*msg++ = htonl(argp->excl);
	msg = xdr_encode_nlm_lock(msg, &argp->lock);
	*msg++ = htonl(argp->reclaim);
	*msg++ = htonl(argp->state);
	return msg;
}

static u32 *
xdrsvc_nlm_lockargs(struct svcreq *req, u32 *msg, nlm_lockargs *argp)
{
	argp->cookie = ntohl(*msg++);
	argp->block = ntohl(*msg++);
	argp->excl = ntohl(*msg++);
	msg = xdr_decode_nlm_lock(msg, &argp->lock);
	argp->reclaim = ntohl(*msg++);
	argp->state = ntohl(*msg++);
	return msg;
}

static u32 *
xdr_nlm_unlockargs(struct rpcreq *req, u32 *msg, nlm_unlockargs *argp)
{
	*msg++ = htonl(argp->cookie);
	return xdr_encode_nlm_lock(msg, &argp->lock);
}

static u32 *
xdrsvc_nlm_unlockargs(struct svcreq *req, u32 *msg, nlm_unlockargs *argp)
{
	argp->cookie = ntohl(*msg++);
	return xdr_encode_nlm_lock(msg, &argp->lock);
}

static u32 *
xdr_nlm_cancargs(struct rpcreq *req, u32 *msg, nlm_cancargs *argp)
{
	*msg++ = htonl(argp->cookie);
	*msg++ = htonl(argp->block);
	*msg++ = htonl(argp->excl);
	return xdr_encode_nlm_lock(msg, &argp->lock);
}

static u32 *
xdrsvc_nlm_cancargs(struct svcreq *req, u32 *msg, nlm_cancargs *argp)
{
	argp->cookie = ntohl(*msg++);
	argp->block = ntohl(*msg++);
	argp->excl = ntohl(*msg++);
	return xdr_decode_nlm_lock(msg, &argp->lock);
}

static u32 *
xdr_nlm_res(struct rpcreq *req, u32 *msg, nlm_res *resp)
{
	resp->cookie = ntohl(*msg++);
	resp->status = ntohl(*msg++);
	return msg;
}

static u32 *
xdrsvc_nlm_res(struct svcreq *req, u32 *msg, nlm_res *resp)
{
	*msg++ = htonl(resp->cookie);
	*msg++ = htonl(resp->status);
	return msg;
}

static u32 *
xdr_nlm_testres(struct rpcreq *req, u32 *msg, nlm_testres *resp)
{
	resp->cookie = ntohl(*msg++);
	resp->status = ntohl(*msg++);
	if (resp->status == NLM_LCK_DENIED)
		msg = xdr_decode_nlm_holder(msg, &resp->holder);
	return msg;
}

static u32 *
xdrsvc_nlm_testres(struct svcreq *req, u32 *msg, nlm_testres *resp)
{
	*msg++ = htonl(resp->cookie);
	*msg++ = htonl(resp->status);
	if (resp->status == NLM_LCK_DENIED)
		msg = xdr_encode_nlm_holder(msg, &resp->holder);
	return msg;
}

/*
 * NLM async procedures
 */
static u32 *
xdr_nlm_null(struct rpcreq *req, u32 *msg, void *dummy)
{
	return msg;
}

static u32 *
xdrsvc_nlm_null(struct svcreq *req, u32 *msg, void *dummy)
{
	return msg;
}

static u32 *
xdr_nlm_nullres(struct rpcreq *req, u32 *msg, void *dummy)
{
	return msg;
}

static u32 *
xdrsvc_nlm_nullres(struct svcreq *req, u32 *msg, void *dummy)
{
	return msg;
}

#define RPC_PROTOCOL		NLM_PROTOCOL
#define RPC_PROT_NAME		nlm
#define RPC_SERVER_PROTO
#include <xdr/xdrdefs.h>
