/*
 * rpc.c	RPC header generation, XDR, etc.
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <sys/types.h>
#include <sys/param.h>
#include <unistd.h>
#include <errno.h>
#include "nfshell.h"

static u32	xid = 0;
u32		rpcstat;
int		auth_uid = DEFAULT_UID;
int		auth_gid = DEFAULT_GID;

/*
 * Encode header for RPC call
 */
u32 *
rpc_encode_header(u32 *msg, u32 prog, u32 vers, u32 proc)
{
	if (!xid)
		xid = getpid() << 16;

	*msg++ = xid++;		/* XID */
	*msg++ = 0;		/* CALL */
	*msg++ = htonl(2);	/* RPC version */
	*msg++ = htonl(prog);
	*msg++ = htonl(vers);
	*msg++ = htonl(proc);
	*msg++ = htonl(1);	/* AUTH_UNIX */
	*msg++ = htonl(20);	/* cred len */
	*msg++ = htonl(jiffies() / HZ);
	*msg++ = 0;		/* no machine name */
	*msg++ = htonl(auth_uid);
	*msg++ = htonl(auth_gid);
	*msg++ = 0;		/* empty gid list */
	*msg++ = 0;		/* NULL verf */
	*msg++ = 0;		/* zero length verf */
	return msg;
}

/*
 * Decode header of RPC message
 */
unsigned int
rpc_decode_header(struct svcreq *req, u32 **msgp)
{
	u32		*msg = req->rq_msg;
	u32		auth, alen, reject_stat, auth_error = 0;

	req->rq_xid = *msg++;		/* XID */
	msg++;				/* CALL already checked */

	if (ntohl(*msg++) != 2) {	/* VERSION */
		reject_stat = 0;	/* VERSMISMATCH */
		goto msg_rejected;
	}

	req->rq_prog = ntohl(*msg++);
	req->rq_vers = ntohl(*msg++);
	req->rq_proc = ntohl(*msg++);

	/* credentials */
	auth = ntohl(*msg++);
	alen = ntohl(*msg++);
	if (auth != 0 && auth != 1) {
		reject_stat = 1;	/* AUTHERROR */
		auth_error = 1;		/* BADCRED */
		goto msg_rejected;
	}
	msg += ((alen + 3) >> 2);

	/* verifier */
	auth = ntohl(*msg++);
	alen = ntohl(*msg++);
	if (auth != 0 && auth != 1) {
		reject_stat = 1;	/* AUTHERROR */
		auth_error = 3;		/* BADVERF */
		goto msg_rejected;
	}
	msg += ((alen + 3) >> 2);

	*msgp = msg;
	return 0;

msg_rejected:
	msg = req->rq_rply;
	*msg++ = req->rq_xid;		/* XID */
	*msg++ = htonl(1);		/* REPLY */
	*msg++ = htonl(1);		/* REJECTED */
	*msg++ = htonl(reject_stat);
	*msg++ = htonl(auth_error);

	return 1;
}

/*
 * Encode header of RPC reply
 */
u32 *
rpc_encode_reply(struct svcreq *req, u32 *msg, u32 status)
{
	*msg++ = req->rq_xid;		/* XID */
	*msg++ = htonl(1);		/* REPLY */
	*msg++ = 0;			/* ACCEPTED */
	*msg++ = 0;			/* NULL verf */
	*msg++ = 0;
	*msg++ = htonl(status);		/* reply status */

	return msg;
}

/*
 * Decode header of RPC reply
 */
unsigned int
rpc_decode_reply(struct rpcreq *req, u32 **msgp)
{
	u32	*msg = *msgp;
	int	len;

	req->rq_rpcstat = RPC_FAILED;	/* default error */

	msg++;				/* XID */
	if (ntohl(*msg++) != 1) {	/* direction */
		req->rq_rpcstat = RPC_CANTDECODERES;
	} else
	if (ntohl(*msg++) == 0) {	/* reply_status */
		/* accepted */
		msg += 1;		/* skip verf flavor */
		len  = ntohl(*msg++);	/* verf len */
		msg += (len + 3) >> 2;

		switch (ntohl(*msg++)) {/* accept_status */
		case 0: req->rq_rpcstat = RPC_SUCCESS;
			break;
		case 1: req->rq_rpcstat = RPC_PROGUNAVAIL;
			break;
		case 2: req->rq_rpcstat = RPC_PROGVERSMISMATCH;
			break;
		case 3: req->rq_rpcstat = RPC_PROCUNAVAIL;
			break;
		case 4: req->rq_rpcstat = RPC_CANTDECODEARGS;
			break;
		case 5: req->rq_rpcstat = RPC_SYSTEMERROR;
			break;
		}
	} else {
		/* rejected */
		switch (ntohl(*msg++)) {/* reject_status */
		case 0: req->rq_rpcstat = RPC_VERSMISMATCH;
			break;
		case 1: req->rq_rpcstat = RPC_AUTHERROR;
			break;
		}
	}

	*msgp = msg;
	if (req->rq_rpcstat == RPC_SUCCESS)
		return (msg - req->rq_rply) << 2;

	return 0;
}

const char *
rpc_strerror(u32 stat)
{
	switch (stat) {
	case RPC_SUCCESS:
		return "RPC_SUCCESS {call succeeded}";
	case RPC_CANTENCODEARGS:
		return "RPC_CANTENCODEARGS {can't encode arguments}";
	case RPC_CANTDECODERES:
		return "RPC_CANTDECODERES {can't decode results}";
	case RPC_CANTSEND:
		return "RPC_CANTSEND {failure in sending call}";
	case RPC_CANTRECV:
		return "RPC_CANTRECV {failure in receiving result}";
	case RPC_TIMEDOUT:
		return "RPC_TIMEDOUT {call timed out}";
	case RPC_VERSMISMATCH:
		return "RPC_VERSMISMATCH {rpc versions not compatible}";
	case RPC_AUTHERROR:
		return "RPC_AUTHERROR {authentication error}";
	case RPC_PROGUNAVAIL:
		return "RPC_PROGUNAVAIL {program not available}";
	case RPC_PROGVERSMISMATCH:
		return "RPC_PROGVERSMISMATCH {program version mismatched}";
	case RPC_PROCUNAVAIL:
		return "RPC_PROCUNAVAIL {procedure unavailable}";
	case RPC_CANTDECODEARGS:
		return "RPC_CANTDECODEARGS {decode arguments error}";
	case RPC_SYSTEMERROR:
		return "RPC_SYSTEMERROR {generic other problem}";
	case RPC_UNKNOWNHOST:
		return "RPC_UNKNOWNHOST {unknown host name}";
	case RPC_UNKNOWNPROTO:
		return "RPC_UNKNOWNPROTO {unkown protocol}";
	case RPC_PMAPFAILURE:
		return "RPC_PMAPFAILURE {the pmapper failed in its call}";
	case RPC_PROGNOTREGISTERED:
		return "RPC_PROGNOTREGISTERED {remote program not registered}";
	}
	return "RPC_FAILED {general failure?}";
}
